"""
PyArrow implementation for aggregated feature groups.
"""

from __future__ import annotations

from typing import Any, Set, Type, Union

import pyarrow as pa
import pyarrow.compute as pc

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.compute_frame_work import ComputeFrameWork

from mloda_plugins.compute_framework.base_implementations.pyarrow.table import PyarrowTable
from mloda_plugins.feature_group.experimental.aggregated_feature_group.base import BaseAggregatedFeatureGroup


class PyArrowAggregatedFeatureGroup(BaseAggregatedFeatureGroup):
    """
    PyArrow implementation of aggregated feature group.

    Supports multiple aggregation types in a single class.
    """

    @classmethod
    def compute_framework_rule(cls) -> Union[bool, Set[Type[ComputeFrameWork]]]:
        """Specify that this feature group works with PyArrow."""
        return {PyarrowTable}

    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        """
        Perform aggregations using PyArrow.

        Processes all requested features, determining the aggregation type
        and source feature from each feature name.

        Adds the aggregated results to the input Table.
        """
        # Process each requested feature
        for feature_name in features.get_all_names():
            aggregation_type = cls.get_aggregation_type(feature_name)
            source_feature = cls.get_source_feature(feature_name)

            # Check if source feature exists in the table
            if source_feature not in data.schema.names:
                raise ValueError(f"Source feature '{source_feature}' not found in data")

            if aggregation_type not in cls.AGGREGATION_TYPES:
                raise ValueError(f"Unsupported aggregation type: {aggregation_type}")

            # Apply the appropriate aggregation function
            result = cls._perform_aggregation(data, aggregation_type, source_feature)

            # Create an array with the aggregated result repeated for each row
            repeat_count = data.num_rows
            repeated_result = pa.array([result] * repeat_count)

            # Add the new column to the table
            data = data.append_column(feature_name, repeated_result)

        # Return the modified Table
        return data

    @classmethod
    def _perform_aggregation(cls, data: Any, aggregation_type: str, source_feature: str) -> Any:
        """
        Perform the aggregation using PyArrow compute functions.

        Args:
            data: The PyArrow Table
            aggregation_type: The type of aggregation to perform
            source_feature: The name of the source feature to aggregate

        Returns:
            The result of the aggregation
        """
        # Get the column to aggregate
        column = data.column(source_feature)

        if aggregation_type == "sum":
            return pc.sum(column).as_py()
        elif aggregation_type == "min":
            return pc.min(column).as_py()
        elif aggregation_type == "max":
            return pc.max(column).as_py()
        elif aggregation_type in ["avg", "mean"]:
            return pc.mean(column).as_py()
        elif aggregation_type == "count":
            return pc.count(column).as_py()
        elif aggregation_type == "std":
            return pc.stddev(column).as_py()
        elif aggregation_type == "var":
            return pc.variance(column).as_py()
        elif aggregation_type == "median":
            # PyArrow doesn't have a direct median function
            # We can approximate it using quantile with q=0.5
            # quantile returns an array, so we need to extract the first value
            result = pc.quantile(column, q=0.5)
            return result[0].as_py()
        else:
            raise ValueError(f"Unsupported aggregation type: {aggregation_type}")



import logging
import os
from typing import Any, Optional, Set

from mloda_core.abstract_plugins.abstract_feature_group import AbstractFeatureGroup
from mloda_core.abstract_plugins.components.feature import Feature
from mloda_core.abstract_plugins.components.feature_name import FeatureName
from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.components.options import Options
from mloda_plugins.feature_group.experimental.llm.llm_api.gemini import GeminiRequestLoop
from mloda_plugins.feature_group.input_data.read_context_files import ConcatenatedFileContent
from mloda_plugins.feature_group.experimental.default_options_key import DefaultOptionKeys

logger = logging.getLogger(__name__)


class LLMFileSelector(AbstractFeatureGroup):
    """
    A feature group that uses an LLM to select relevant files from a directory.
    """

    def input_features(self, options: Options, feature_name: FeatureName) -> Optional[Set[Feature]]:
        prompt = options.get("prompt")
        if not prompt:
            raise ValueError("Prompt is required for LLMFileSelector")

        target_folder = options.get("target_folder")
        if not target_folder:
            raise ValueError("Target folder is required for LLMFileSelector")

        disallowed_files = list(options.get("disallowed_files")) if options.get("disallowed_files") else ["__init__.py"]

        file_type = options.get("file_type")

        llm_feature = Feature(
            name=GeminiRequestLoop.get_class_name(),
            options={
                "model": "gemini-2.0-flash-exp",  # Choose your desired model
                "prompt": prompt,
                DefaultOptionKeys.mloda_source_feature: frozenset([ConcatenatedFileContent.get_class_name()]),
                "target_folder": frozenset(target_folder),
                "disallowed_files": frozenset(disallowed_files),
                "file_type": file_type,
            },
        )
        return {llm_feature}

    @classmethod
    def validate_input_features(cls, data: Any, features: FeatureSet) -> Optional[bool]:
        if GeminiRequestLoop.get_class_name() not in data.columns:
            raise ValueError(f"Feature {GeminiRequestLoop.get_class_name()} not found in input data.")

        for str_paths in data[GeminiRequestLoop.get_class_name()].values:
            paths = str_paths.split(",")
            for path in paths:
                if "\n" in path:
                    raise ValueError(f"File path {path} contains a newline character.")

                if not os.path.exists(path):
                    raise ValueError(f"File <{path}> does not exist.")

        return True

    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        data = data.rename(
            columns={GeminiRequestLoop.get_class_name(): LLMFileSelector.get_class_name()}, inplace=False
        )
        return data

    @classmethod
    def validate_output_features(cls, data: Any, features: FeatureSet) -> Optional[bool]:
        for str_paths in data[cls.get_class_name()].values:
            paths = str_paths.split(",")
            for path in paths:
                if "\n" in path:
                    raise ValueError(f"File path {path} contains a newline character.")

                if not os.path.exists(path):
                    raise ValueError(f"File <{path}> does not exist.")
        return True



from typing import Any, Dict, Set, Tuple, Type
from mloda_core.abstract_plugins.components.merge.base_merge_engine import BaseMergeEngine
from mloda_plugins.compute_framework.base_implementations.pandas.pandas_merge_engine import PandasMergeEngine
import pyarrow as pa
from mloda_core.abstract_plugins.components.cfw_transformer import ComputeFrameworkTransformMap
from mloda_core.abstract_plugins.components.feature_name import FeatureName
from mloda_core.abstract_plugins.compute_frame_work import ComputeFrameWork
from mloda_core.filter.filter_engine import BaseFilterEngine
from mloda_plugins.compute_framework.base_implementations.pandas.pandas_filter_engine import PandasFilterEngine

try:
    import pandas as pd
except ImportError:
    pd = None


class PandasDataframe(ComputeFrameWork):
    @staticmethod
    def pd_dataframe() -> Any:
        if pd is None:
            raise ImportError("Pandas is not installed. To be able to use this framework, please install pandas.")
        return pd.DataFrame

    @staticmethod
    def pd_series() -> Any:
        if pd is None:
            raise ImportError("Pandas is not installed. To be able to use this framework, please install pandas.")
        return pd.Series

    @staticmethod
    def pd_merge() -> Any:
        if pd is None:
            raise ImportError("Pandas is not installed. To be able to use this framework, please install pandas.")
        return pd.merge

    @staticmethod
    def expected_data_framework() -> Any:
        return PandasDataframe.pd_dataframe()

    def transform(
        self,
        data: Any,
        feature_names: Set[str],
        transform_map: ComputeFrameworkTransformMap = ComputeFrameworkTransformMap(),
    ) -> Any:
        transformed_data = self.transform_refactored(data, transform_map)
        if transformed_data is not None:
            return transformed_data

        if isinstance(data, dict):
            """Initial data: Transform dict to table"""
            return self.pd_dataframe().from_dict(data)

        if isinstance(data, self.pd_series()):
            """Added data: Add column to table"""
            if len(feature_names) == 1:
                feature_name = next(iter(feature_names))

                if feature_name in self.data.columns:
                    raise ValueError(f"Feature {feature_name} already exists in the dataframe")

                self.data[feature_name] = data
                return self.data
            raise ValueError(f"Only one feature can be added at a time: {feature_names}")

        raise ValueError(f"Data {type(data)} is not supported by {self.__class__.__name__}")

    def select_data_by_column_names(self, data: Any, selected_feature_names: Set[FeatureName]) -> Any:
        return data[[f.name for f in selected_feature_names]]

    def set_column_names(self) -> None:
        self.column_names = set(self.data.columns)

    @staticmethod
    def get_framework_transform_functions(from_other: bool, other: Any) -> Tuple[Any, Dict[str, Any]]:
        def pyarrow_table_to_pandas_dataframe(data: pd.DataFrame, parameters: Dict[str, Any] = {}) -> pa.Table:
            return pa.Table.to_pandas(data, **parameters)

        def pandas_dataframe_to_pyarrow_table(data: pd.DataFrame, parameters: Dict[str, Any] = {}) -> pa.Table:
            # drop pandas schema metadata
            pyarrow_table = pa.Table.from_pandas(data, **parameters)
            schema = pyarrow_table.schema
            metadata = schema.metadata.copy() if schema.metadata else {}
            metadata.pop(b"pandas", None)
            new_schema = schema.with_metadata(metadata)
            return pa.Table.from_arrays(pyarrow_table.columns, schema=new_schema)

        if other == pa.Table:
            if from_other:
                return pyarrow_table_to_pandas_dataframe, {}
            return pandas_dataframe_to_pyarrow_table, {}

        return None, {"defensive": None}

    def filter_engine(self) -> Type[BaseFilterEngine]:
        return PandasFilterEngine

    def merge_engine(self) -> Type[BaseMergeEngine]:
        return PandasMergeEngine



from typing import Any, Optional

from mloda_plugins.feature_group.experimental.default_options_key import DefaultOptionKeys


class Options:
    """
    Options can be passed into the feature, allowing arbitrary variables to be used.
    This enables configuration:
    - at request time
    - when defining input features of a feature_group.

    At-request options are forwarded to child features. This allows configuring children features by:
    - request feature options
    - defining input features of the feature_group.
    """

    def __init__(self, data: Optional[dict[str, Any]] = None) -> None:
        self.data = data or {}

    def add(self, key: str, value: Any) -> None:
        if key in self.data:
            raise ValueError(f"Key {key} already exists in options.")

        self.data[key] = value

    def __hash__(self) -> int:
        return hash(frozenset(self.data.items()))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Options):
            return False
        return self.data == other.data

    def get(self, key: str) -> Any:
        return self.data.get(key, None)

    def __str__(self) -> str:
        return str(self.data)

    def update_considering_mloda_source(self, other: "Options") -> None:
        """
        Updates the options object with data from another Options object, excluding the mloda_source_feature key.

        The mloda_source_feature key is excluded to preserve the parent feature source, as it is not relevant to the child feature.
        """

        exclude_key = DefaultOptionKeys.mloda_source_feature

        other_data_copy = other.data.copy()
        if exclude_key in other_data_copy and exclude_key in self.data:
            del other_data_copy[exclude_key]

        self.data.update(other_data_copy)



"""
Pandas implementation for aggregated feature groups.
"""

from __future__ import annotations

from typing import Any, Set, Type, Union

from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.compute_frame_work import ComputeFrameWork

from mloda_plugins.compute_framework.base_implementations.pandas.dataframe import PandasDataframe
from mloda_plugins.feature_group.experimental.aggregated_feature_group.base import BaseAggregatedFeatureGroup


class PandasAggregatedFeatureGroup(BaseAggregatedFeatureGroup):
    """
    Pandas implementation of aggregated feature group.

    Supports multiple aggregation types in a single class.
    """

    @classmethod
    def compute_framework_rule(cls) -> Union[bool, Set[Type[ComputeFrameWork]]]:
        """Specify that this feature group works with Pandas."""
        return {PandasDataframe}

    @classmethod
    def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:
        """
        Perform aggregations using Pandas.

        Processes all requested features, determining the aggregation type
        and source feature from each feature name.

        Adds the aggregated results directly to the input DataFrame.
        """
        # Process each requested feature
        for feature_name in features.get_all_names():
            aggregation_type = cls.get_aggregation_type(feature_name)
            source_feature = cls.get_source_feature(feature_name)

            if source_feature not in data.columns:
                raise ValueError(f"Source feature '{source_feature}' not found in data")

            if aggregation_type not in cls.AGGREGATION_TYPES:
                raise ValueError(f"Unsupported aggregation type: {aggregation_type}")

            # Apply the appropriate aggregation function and add result to the DataFrame
            data[feature_name] = cls._perform_aggregation(data, aggregation_type, source_feature)

        # Return the modified DataFrame
        return data

    @classmethod
    def _perform_aggregation(cls, data: Any, aggregation_type: str, source_feature: str) -> Any:
        """
        Perform the aggregation using Pandas.

        Args:
            data: The Pandas DataFrame
            aggregation_type: The type of aggregation to perform
            source_feature: The name of the source feature to aggregate

        Returns:
            The result of the aggregation
        """
        if aggregation_type == "sum":
            return data[source_feature].sum()
        elif aggregation_type == "min":
            return data[source_feature].min()
        elif aggregation_type == "max":
            return data[source_feature].max()
        elif aggregation_type in ["avg", "mean"]:
            return data[source_feature].mean()
        elif aggregation_type == "count":
            return data[source_feature].count()
        elif aggregation_type == "std":
            return data[source_feature].std()
        elif aggregation_type == "var":
            return data[source_feature].var()
        elif aggregation_type == "median":
            return data[source_feature].median()
        else:
            raise ValueError(f"Unsupported aggregation type: {aggregation_type}")



from typing import Any, Dict, Optional, Type, Set, List, Union
from mloda_core.abstract_plugins.abstract_feature_group import AbstractFeatureGroup
from mloda_core.abstract_plugins.components.options import Options
from mloda_core.abstract_plugins.components.feature_name import FeatureName
from mloda_core.abstract_plugins.components.data_access_collection import DataAccessCollection
from mloda_core.abstract_plugins.components.feature_set import FeatureSet
from mloda_core.abstract_plugins.components.data_types import DataType
from mloda_core.abstract_plugins.compute_frame_work import ComputeFrameWork
from mloda_core.abstract_plugins.components.index.index import Index
from mloda_core.abstract_plugins.components.input_data.base_input_data import BaseInputData


class DynamicFeatureGroupCreator:
    """
    A class to dynamically create AbstractFeatureGroup subclasses based on a dictionary of properties.
    """

    _created_classes: Dict[str, Type[AbstractFeatureGroup]] = {}  # Store created classes

    @staticmethod
    def create(
        properties: Dict[str, Any],
        class_name: str = "DynamicFeatureGroup",
        feature_group_cls: Type[AbstractFeatureGroup] = AbstractFeatureGroup,
    ) -> Type[AbstractFeatureGroup]:
        """
        Creates a new AbstractFeatureGroup subclass with the given properties.

        Args:
            properties: A dictionary containing the properties for the new class.
            class_name: The name of the new class.

        Returns:
            A new AbstractFeatureGroup subclass.
        """

        if class_name in DynamicFeatureGroupCreator._created_classes:
            return DynamicFeatureGroupCreator._created_classes[class_name]

        def set_feature_name(self, config: Options, feature_name: FeatureName) -> FeatureName:  # type: ignore
            if "set_feature_name" in properties:
                return properties["set_feature_name"](self, config, feature_name)  # type: ignore
            return feature_name

        def match_feature_group_criteria(  # type: ignore
            cls,
            feature_name: Union[FeatureName, str],
            options: Options,
            data_access_collection: Optional[DataAccessCollection] = None,
        ) -> bool:
            if "match_feature_group_criteria" in properties:
                return properties["match_feature_group_criteria"](cls, feature_name, options, data_access_collection)  # type: ignore
            return super(new_class, cls).match_feature_group_criteria(feature_name, options, data_access_collection)  # type: ignore

        def input_data(cls) -> Optional[BaseInputData]:  # type: ignore
            if "input_data" in properties:
                return properties["input_data"]()  # type: ignore
            return super(new_class, cls).input_data()  # type: ignore

        def validate_input_features(cls, data: Any, features: FeatureSet) -> Optional[bool]:  # type: ignore
            if "validate_input_features" in properties:
                return properties["validate_input_features"](cls, data, features)  # type: ignore
            return super(new_class, cls).validate_input_features(data, features)  # type: ignore

        def calculate_feature(cls, data: Any, features: FeatureSet) -> Any:  # type: ignore
            if "calculate_feature" in properties:
                return properties["calculate_feature"](cls, data, features)
            return super(new_class, cls).calculate_feature(data, features)  # type: ignore

        def validate_output_features(cls, data: Any, features: FeatureSet) -> Optional[bool]:  # type: ignore
            if "validate_output_features" in properties:
                return properties["validate_output_features"](cls, data, features)  # type: ignore
            return super(new_class, cls).validate_output_features(data, features)  # type: ignore

        def artifact(cls) -> Optional[Type[Any]]:  # type: ignore
            if "artifact" in properties:
                return properties["artifact"]()  # type: ignore
            return super(new_class, cls).artifact()  # type: ignore

        def compute_framework_rule(cls) -> Union[bool, Set[Type[ComputeFrameWork]]]:  # type: ignore
            if "compute_framework_rule" in properties:
                return properties["compute_framework_rule"]()  # type: ignore
            return super(new_class, cls).compute_framework_rule()  # type: ignore

        def return_data_type_rule(cls, feature: Any) -> Optional[DataType]:  # type: ignore
            if "return_data_type_rule" in properties:
                return properties["return_data_type_rule"](cls, feature)  # type: ignore
            return super(new_class, cls).return_data_type_rule(feature)  # type: ignore

        def input_features(self, options: Options, feature_name: FeatureName) -> Optional[Set[Any]]:  # type: ignore
            if "input_features" in properties:
                return properties["input_features"](self, options, feature_name)  # type: ignore
            return super(new_class, self).input_features(options, feature_name)  # type: ignore

        def index_columns(cls) -> Optional[List[Index]]:  # type: ignore
            if "index_columns" in properties:
                return properties["index_columns"]()  # type: ignore
            return super(new_class, cls).index_columns()  # type: ignore

        def supports_index(cls, index: Index) -> Optional[bool]:  # type: ignore
            if "supports_index" in properties:
                return properties["supports_index"](cls, index)  # type: ignore
            return super(new_class, cls).supports_index(index)  # type: ignore

        new_class = type(
            class_name,
            (feature_group_cls,),
            {
                "set_feature_name": set_feature_name,
                "match_feature_group_criteria": classmethod(match_feature_group_criteria),
                "input_data": classmethod(input_data),
                "validate_input_features": classmethod(validate_input_features),
                "calculate_feature": classmethod(calculate_feature),
                "validate_output_features": classmethod(validate_output_features),
                "artifact": classmethod(artifact),
                "compute_framework_rule": classmethod(compute_framework_rule),
                "return_data_type_rule": classmethod(return_data_type_rule),
                "input_features": input_features,
                "index_columns": classmethod(index_columns),
                "supports_index": classmethod(supports_index),
            },
        )
        DynamicFeatureGroupCreator._created_classes[class_name] = new_class
        return new_class



import os
from typing import Any

from mloda_plugins.feature_group.experimental.llm.tools.base_tool import BaseTool
from mloda_plugins.feature_group.experimental.llm.tools.tool_data_classes import (
    ToolFunctionDeclaration,
    InputDataObject,
)


class AdjustFileTool(BaseTool):
    @classmethod
    def tool_declaration(cls) -> ToolFunctionDeclaration:
        return cls.build_tool_declaration(
            name="AdjustFileTool",
            description="AdjustFileTool replaces content in an existing file.",
            parameters=[
                InputDataObject(name="file_path", type="str", description="The full path of the file to modify."),
                InputDataObject(name="old_content", type="str", description="The exact text to replace."),
                InputDataObject(
                    name="new_content", type="str", description="The text to replace the old content with."
                ),
            ],
            required=["file_path", "old_content", "new_content"],
        )

    @classmethod
    def execute(cls, **kwargs: Any) -> str:
        file_path = kwargs.get("file_path")
        old_content = kwargs.get("old_content")
        new_content = kwargs.get("new_content")

        if file_path is None or old_content is None or new_content is None:
            raise ValueError("Both 'file_path', 'old_content', and 'new_content' parameters are required")

        if not os.path.exists(file_path):
            raise FileNotFoundError(f"The file '{file_path}' does not exist")

        with open(file_path, "r") as f:
            content = f.read()

        _saved_content = content

        new_file_content = content.replace(old_content, new_content)

        with open(file_path, "w") as f:
            f.write(new_file_content)

        if _saved_content == new_file_content:
            return f"File '{file_path}' was not modified. No occurrences of '{old_content}' were found. NO CHANGES WERE MADE."

        return f"Successfully replaced all occurrences of '{old_content}' with '{new_content}' in file '{file_path}'."

    @classmethod
    def create_result_string(cls, result: str, **kwargs: Any) -> str:
        return result



from copy import deepcopy
from typing import Any, Tuple

from mloda_plugins.feature_group.input_data.read_file import ReadFile
from pyarrow import fs as pyarrow_fs

from mloda_core.abstract_plugins.components.feature_set import FeatureSet

try:
    import pandas as pd
except ImportError:
    pd = None


class TextFileReader(ReadFile):
    @classmethod
    def suffix(cls) -> Tuple[str, ...]:
        return (".text",)

    @classmethod
    def load_data(cls, data_access: Any, features: FeatureSet) -> Any:
        if pd is None:
            raise ImportError("pandas is not installed.")

        file_path = deepcopy(features.get_options_key(cls.__name__))
        local_fs = pyarrow_fs.LocalFileSystem()

        with local_fs.open_input_file(file_path) as file:
            content = file.read().decode("utf-8")

        if features.get_options_key("AddPathToContent"):
            file_path = file_path.strip("\n")
            content = f"The file path of the following file is {file_path} : {content}"

        data = {cls.get_class_name(): [content]}

        return pd.DataFrame(data)


class PyFileReader(TextFileReader):
    @classmethod
    def suffix(cls) -> Tuple[str, ...]:
        return (".py",)



from copy import copy
from dataclasses import asdict
import os
import time
from typing import Any, Dict, List, Tuple, Union


from mloda_core.abstract_plugins.components.feature_set import FeatureSet

from mloda_plugins.feature_group.experimental.llm.llm_api.llm_base_request import LLMBaseApi
from mloda_plugins.feature_group.experimental.llm.llm_api.request_loop import RequestLoop
from mloda_plugins.feature_group.experimental.llm.tools.tool_collection import ToolCollection
from mloda_plugins.feature_group.experimental.llm.tools.tool_data_classes import ToolFunctionDeclaration

try:
    import anthropic
    from anthropic.types.tool_use_block import ToolUseBlock
except ImportError:
    anthropic, ToolUseBlock = None, None

try:
    import pandas as pd
except ImportError:
    pd = None


import logging

logger = logging.getLogger(__name__)


def python_type_to_claude_type(python_type: str) -> str:
    """Converts Python type strings to Claude API type strings."""
    type_mapping = {
        "float": "number",
        "int": "integer",
        "str": "string",
        "bool": "boolean",
        "number": "number",
    }
    return type_mapping.get(python_type, "string")  # Default to string if not found


def parse_tool_function_for_claude(function_declaration: ToolFunctionDeclaration) -> Dict[str, Any]:
    """Parses a ToolFunctionDeclaration into a dict formatted for Claude function calling.

    The output will have the following structure compatible with Anthropic's API:
    {
        "name": <function name>,
        "description": <function description>,
        "input_schema": {
            "type": "object",
            "properties": {
                <param_name>: {
                    "type": <claude json schema type>,
                    "description": <param description>
                },
                ...
            },
            "required": [<list of required param names>]
        }
    }
    """
    if not isinstance(function_declaration, ToolFunctionDeclaration):
        raise ValueError(f"Tool function {function_declaration} does not return a ToolFunctionDeclaration.")

    # Build the parameters schema
    input_schema = {
        "type": "object",
        "properties": {},
        "required": function_declaration.required,
    }

    for param in function_declaration.parameters:
        param_dict = asdict(param)
        input_schema["properties"][param_dict["name"]] = {  # type: ignore
            "type": python_type_to_claude_type(param_dict["type"]),
            "description": param_dict["description"],
        }

    # Construct the final tool structure
    tool_dict = {
        "name": function_declaration.name,
        "description": function_declaration.description,
        "input_schema": input_schema,
    }

    return tool_dict


class ClaudeAPI(LLMBaseApi):
    @classmethod
    def request(
        cls,
        model: str,
        prompt: Union[str, List[Dict[str, str]]],
        model_parameters: Dict[str, Any],
        tools: ToolCollection | None,
    ) -> Any:
        if isinstance(prompt, str):
            raise ValueError("Claude requires a list of messages, not a single string.")

        try:
            claude_client = cls._setup_model_if_needed(model)
            if claude_client is not None:
                _tools = cls.parse_tools(tools)
                result = cls.generate_response(claude_client, model, prompt, _tools)
                return result
        except Exception as e:
            logger.error(f"Error during Claude request: {e}")
            raise

        raise ValueError("Claude model is not set.")

    @classmethod
    def _setup_model_if_needed(cls, model: str) -> Any:
        api_key = os.environ.get("CLAUDE_API_KEY")
        if not api_key:
            raise ValueError("CLAUDE_API_KEY environment variable is not set.")

        if anthropic is None:
            raise ImportError("Please install the anthropic package to use Claude.")

        claude_client = anthropic.Anthropic(api_key=api_key)
        return claude_client

    @classmethod
    def parse_tools(cls, tools: ToolCollection | None) -> List[Dict[str, Any]]:
        """Parses all tools in the ToolCollection for Claude."""
        parsed_tools: List[Dict[str, Any]] = []
        if tools is None:
            return parsed_tools

        for _, tool in tools.get_all_tools().items():
            parsed_tools.append(parse_tool_function_for_claude(tool))

        return parsed_tools

    @classmethod
    def handle_response(
        cls, response: Any, features: FeatureSet, tools: ToolCollection | None
    ) -> Tuple[List[Dict[str, str]], str]:
        responses = []
        used_tool = ""

        if hasattr(response, "content"):
            for block in response.content:
                if hasattr(block, "text") and block.text:
                    responses.append({"text": block.text})

                if isinstance(block, ToolUseBlock):
                    if tools is None:
                        raise ValueError("Tools are not set.")

                    tool_dict = {"name": block.name, "args": block.input}
                    tool_result = cls._execute_tools([tool_dict], features, tools)
                    if tool_result:
                        responses.append({"tool": tool_result})
                        used_tool += tool_result
        else:
            logger.warning(f"Response has unexpected structure: {response}")
            return str(response), ""  # type: ignore

        return responses, used_tool

    def generate_response(
        client: Any,
        model: str,
        messages: List[Dict[str, str]],
        tools: Any,
        max_retries: int = 5,
        initial_retry_delay: int = 10,
        max_retry_delay: int = 60,
    ) -> Any:
        """
        Generates content from Claude with retry logic for rate limits.
        """
        # Override defaults with environment variables if present
        max_retries = int(os.environ.get("CLAUDE_MAX_RETRIES", str(max_retries)))
        initial_retry_delay = int(os.environ.get("CLAUDE_INITIAL_RETRY_DELAY", str(initial_retry_delay)))
        max_retry_delay = int(os.environ.get("CLAUDE_MAX_RETRY_DELAY", str(max_retry_delay)))

        retry_attempt = 0
        while retry_attempt <= max_retries:
            try:
                message_params = {
                    "model": model,
                    "messages": messages,
                    "max_tokens": 1000,
                }

                if tools:
                    message_params["tools"] = tools

                result = client.messages.create(**message_params)
                return result
            except Exception as e:
                is_rate_limit_error = False
                if hasattr(e, "status_code") and e.status_code == 429:
                    is_rate_limit_error = True

                if is_rate_limit_error:
                    retry_attempt += 1
                    if retry_attempt > max_retries:
                        print(f"Maximum retries ({max_retries}) reached for CLAUDE. Raising exception.")
                        raise
                    delay = min(initial_retry_delay * (2 ** (retry_attempt - 1)), max_retry_delay)
                    print(
                        f"Rate limit hit for CLAUDE. Retrying in {delay:.2f} seconds (Attempt {retry_attempt}/{max_retries})."
                    )
                    time.sleep(delay)
                else:
                    print(f"An unexpected error occurred during CLAUDE request: {e}")
                    raise

        raise Exception(f"Maximum retries ({max_retries}) reached for CLAUDE without a successful response.")


class ClaudeRequestLoop(RequestLoop):
    @classmethod
    def api(cls) -> Any:
        return ClaudeAPI

    @classmethod
    def initial_prompt_message(cls, messages: Any, initial_prompt: str) -> Tuple[Any, Any]:
        if not messages:
            messages = [{"role": "user", "content": initial_prompt}]
            _messages = copy(messages)
        else:
            _messages = copy(messages)
            _messages.append({"role": "user", "content": cls.add_final_part_of_prompt()})

        return messages, _messages

    @classmethod
    def add_tool_response_to_messages(cls, messages: Any, response: str) -> Any:
        messages.append({"role": "user", "content": response})
        return messages

    @classmethod
    def add_text_response_to_messages(cls, messages: Any, response: str) -> Any:
        messages.append({"role": "assistant", "content": response})
        return messages



from typing import Any, Dict, Type
from enum import Enum, auto


class CFWTransformEnums(Enum):
    FUNCTION = auto()
    PARAMETERS = auto()


class ComputeFrameworkTransformMap:
    def __init__(self) -> None:
        self.transform_map: Dict[Any, Any] = {}

    def validate_input(
        self,
        source_compute_frame_work: Type[Any],
        target_compute_frame_work: Type[Any],
        parameters: Dict[str, Any],
    ) -> None:
        if source_compute_frame_work not in self.transform_map:
            raise ValueError(
                f"Source compute framework {source_compute_frame_work} not found in transform map. Thus, we cannot transform to {target_compute_frame_work}."
            )

        if target_compute_frame_work not in self.transform_map[source_compute_frame_work]:
            raise ValueError(
                f"Target compute framework {target_compute_frame_work} not found in transform map for source compute framework {source_compute_frame_work}."
            )

        for parameter, value in parameters.items():
            if self.transform_map[source_compute_frame_work][target_compute_frame_work].get("parameter") is None:
                raise ValueError(
                    f"Parameter {parameter} is not allowed for transformation from {source_compute_frame_work} to {target_compute_frame_work}."
                )

    def validate_output(
        self, source_compute_frame_work: Type[Any], target_compute_frame_work: Type[Any], result_data: Any
    ) -> None:
        if not isinstance(result_data, target_compute_frame_work):
            raise ValueError(
                f"Transformation from {source_compute_frame_work} to {target_compute_frame_work} failed as result is of type {type(result_data)}."
            )

    def transform(
        self,
        data: Any,
        source_compute_frame_work: Type[Any],
        target_compute_frame_work: Type[Any],
        parameters: Dict[str, Any],
    ) -> Any:
        self.validate_input(source_compute_frame_work, target_compute_frame_work, parameters)

        func = self.transform_map[source_compute_frame_work][target_compute_frame_work][CFWTransformEnums.FUNCTION]
        result_data = func(data=data, **parameters)

        self.validate_output(source_compute_frame_work, target_compute_frame_work, result_data)

        return result_data

    def add_transformation(
        self,
        source_compute_frame_work: Any,
        target_cfw: Type[Any],
        function: Any,
        allowed_parameters: Dict[str, Any] = {},
    ) -> None:
        if source_compute_frame_work not in self.transform_map:
            self.transform_map[source_compute_frame_work] = {}

        self.transform_map[source_compute_frame_work][target_cfw] = {
            CFWTransformEnums.FUNCTION: function,
            CFWTransformEnums.PARAMETERS: allowed_parameters,
        }



"""
Base implementation for aggregated feature groups.
"""

from __future__ import annotations

from typing import Any, Optional, Set, Union

from mloda_core.abstract_plugins.abstract_feature_group import AbstractFeatureGroup
from mloda_core.abstract_plugins.components.feature import Feature
from mloda_core.abstract_plugins.components.feature_name import FeatureName
from mloda_core.abstract_plugins.components.options import Options


class BaseAggregatedFeatureGroup(AbstractFeatureGroup):
    """
    Base class for all aggregated feature groups.

    Follows naming convention: {aggregation_type}_aggr_{source_feature_name}
    Examples: sum_aggr_sales, avg_aggr_temperature, max_aggr_price
    """

    # Define supported aggregation types
    AGGREGATION_TYPES = {
        "sum": "Sum of values",
        "min": "Minimum value",
        "max": "Maximum value",
        "avg": "Average (mean) of values",
        "mean": "Average (mean) of values",
        "count": "Count of non-null values",
        "std": "Standard deviation of values",
        "var": "Variance of values",
        "median": "Median value",
    }

    def input_features(self, options: Options, feature_name: FeatureName) -> Optional[Set[Feature]]:
        """Extract source feature from the aggregated feature name."""
        source_feature = self.get_source_feature(feature_name.name)
        return {Feature(source_feature)}

    @classmethod
    def get_aggregation_type(cls, feature_name: str) -> str:
        """Extract the aggregation type from the feature name."""
        parts = feature_name.split("_aggr_")
        if len(parts) != 2 or not parts[0]:
            raise ValueError(f"Invalid aggregated feature name format: {feature_name}")
        if not parts[1]:
            raise ValueError(f"Invalid aggregated feature name format: {feature_name}")
        return parts[0]

    @classmethod
    def get_source_feature(cls, feature_name: str) -> str:
        """Extract the source feature name from the aggregated feature name."""
        parts = feature_name.split("_aggr_")
        if len(parts) != 2 or not parts[1]:
            raise ValueError(f"Invalid aggregated feature name format: {feature_name}")
        if not parts[0]:
            raise ValueError(f"Invalid aggregated feature name format: {feature_name}")
        return parts[1]

    @classmethod
    def match_feature_group_criteria(
        cls,
        feature_name: Union[FeatureName, str],
        options: Options,
        data_access_collection: Optional[Any] = None,
    ) -> bool:
        """Check if feature name matches the expected pattern and aggregation type."""
        if isinstance(feature_name, FeatureName):
            feature_name = feature_name.name

        try:
            agg_type = cls.get_aggregation_type(feature_name)
            return cls._supports_aggregation_type(agg_type)
        except ValueError:
            return False

    @classmethod
    def _supports_aggregation_type(cls, aggregation_type: str) -> bool:
        """Check if this feature group supports the given aggregation type."""
        return aggregation_type in cls.AGGREGATION_TYPES

    @classmethod
    def _perform_aggregation(cls, data: Any, aggregation_type: str, source_feature: str) -> Any:
        """
        Method to perform the aggregation. Should be implemented by subclasses.

        Args:
            data: The input data
            aggregation_type: The type of aggregation to perform
            source_feature: The name of the source feature to aggregate

        Returns:
            The result of the aggregation
        """
        raise NotImplementedError(f"_perform_aggregation not implemented in {cls.__name__}")
